{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name colorize2}
{@description Create an additional color form (dynamic range) compression}

{$I directives.pas}
uses simage,getparam;
var par                 : tparams;
    cc                  : CSImg_Config;
    fnu,fnc,fno         : ansistring;
    ra                  : TAAStr;
    fc,fc1,fu,fm        : TSImg_FPImage;
    mi                  : CSImg_MetaInfo;
    c,i                 : longint;
    g                   : double;
    r                   : array of double;
    lo                  : boolean;

begin
    with par do
    begin
        init;
        paramusage:='<infile1> [<infile2> [<outfile>]]';
        fnu:='';        defparam('<infile1>','Uncompressed input image', @fnu,tpastr,rapdreq);
        fnc:='';        defparam('<infile2>','Compressed input image (default:stdin)', fnc);
        fno:='';        defparam('<outfile>','Destination image (default: stdout)',fno);
        c:=0;           defparam('-c','Generate output by inserting the additional colors into compressed image. First index is 1.\\'+
                                      'If out of range, output is the additional colors.',c);
        g:=0;           defparam('-g','Compression ratio dependency. Default: 0',g);
        ra.size:=0;     defparam('-r','Regularization parameter qr as quantile, see -a. Reduces brightness of additional color in dark regions. Default: 0.75\\'+
                                      'Can be specified multiple times to produce more than one color',ra);
        lo:=false;      defparam('-rgb','Use rgb optimized luminance calculation in insert mode (-c)',lo);
                        defparam('-o','Destination image',fno); deffmt('-o','<outfile>');
                        defparam('-h','Help',helpf);
        run;

        cc:=CSImg_Config.create(SImg_Config_Default);
        cc.debug(cc.name+': values: fnu='+fnu+' fnc='+fnc+' fno='+fno+' c='+int2str(c)+' g='+real2str(g)+' ra.size='+int2str(ra.size)+' lo='+boolean2str(lo));

        if ra.size<1 then
        begin
            setlength(r,1);
            r[0]:=0.75;
        end
        else begin
            setlength(r,ra.size);
            for i:=0 to ra.size-1 do
                if not val2(ra.aa[i],r[i]) then cc.error(cc.name+': Number expected after -r, got "'+ra.aa[i]+'"');
        end;

        { uncompressed input images }
        mi:=CSImg_MetaInfo.create();
        fu.pt:=nil;
        cc.Info(cc.name+': Reading uncompressed input image '+fnu);
        SImg_ReadImageFile(cc, fnu, fu, mi);
        if fu.nc>1 then SImg_AddColors(cc, fu,fu);

        { compressed input images }
        fc.pt:=nil; fc1.pt:=nil; fm.pt:=nil;
        cc.Info(cc.name+': Reading compressed input image '+fnc);
        SImg_ReadImageFile(cc, fnc, fc, mi);
        if fc.nc>1 then
        begin
            SImg_AddColors(cc, fc,fc1);
            SImg_MaxColor(cc, fc,fm);
        end
        else begin
            fc1:=fc;
            fm:=fc;
        end;

        { new color }
        cc.Info(cc.name+': Calculating new color');
        SImg_Colorize2(cc, fu,fc1,fm, fu, g,1,r);
        if fc1.pt<>fc.pt then SImg_Free(cc, fc1);
        if fm.pt <>fc.pt then SImg_Free(cc, fm);

        { writing output image }
        c-=1;
        if (c<0) or (c>fc.nc) then
        begin
            cc.info(cc.name+': Write output image '+fno);
            SImg_WriteImageFile(cc, fno, fu, mi);
        end
        else begin
            cc.info(cc.name+': Generating and writing output image '+fno);
            SImg_InsertColors(cc, fc,fu,fc, c, lo);
            SImg_WriteImageFile(cc, fno, fc, mi);
        end;
        SImg_Free(cc, fu);
        SImg_Free(cc, fc);
        mi.destroy;
        done;
    end;

    cc.destroy;
end.

