{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name log}
{@description Logarithmic brightness scaling of an image}

{$I directives.pas}
uses simage,getparam,sysutils;
var par         : tparams;
    cc          : CSImg_Config;
    img         : TSImg_FPImage;
    fi,fo       : string;
    l,r,s       : double;
    mi          : CSImg_MetaInfo;

begin
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]]';
        fi:='';     defparam('<infile>','Source image (default: stdin)',fi);
        fo:='';     defparam('<outfile>','Destination image (default: stdout)',fo);
                    defparam('-h','Help',helpf);
        l:=0.10;    defparam('-l','Black reference as quantile (default: 10%)',l);
        r:=1;       defparam('-r','Regularization relative to noise floor (default: 1)',r);
                    defparam('-o','Destination image',fo); deffmt('-o','<outfile>');
        run;

        cc:=CSImg_Config.create(SImg_Config_Default);
        cc.debug(cc.name+': values: source='+fi+' dest='+fo+' l='+real2str(l)+' r='+real2str(r));

        mi:=CSImg_MetaInfo.create();
        cc.Info(cc.name+': Reading input image '+fi);
        SImg_ReadImageFile(cc,fi,img,mi);
        mi.deleteRaw;
        if not mi.get('NOISEFLR',s) then
        begin
            s:=SImg_NoiseFloor(cc,img,2.5);
            mi.add('NOISEFLR', real2str(s,10), 'Noise floor');
        end;

        cc.Info(cc.name+': Logarithmic brightness scaling');
        SImg_Log(cc, img, l,s*r);

        cc.Info(cc.name+': Write output image '+fo);
        SImg_WriteImageFile(cc,fo,img,mi);
        SImg_Free(cc,img);

        done;
    end;

    cc.destroy;
end.
