{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name ppmsources}
{@description Calculate (point) sources from catalog}

{$I directives.pas}
uses simage,getparam,sysutils,dateutils;
var par                 : tparams;
    cc                  : CSImg_Config;
    dbd,ofn             : ansistring;
    n,year              : longint;
    ra,dec,width,height : double;
    rot,scale,proj      : double;
    mi                  : CSImg_MetaInfo;
    pl                  : TSImg_RDImage;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='[<db dir> [<outfile>]]';
        dbd:='';      defparam('<infile>','Directory of the database (default: .)',dbd);
        ofn:='';      defparam('<outfile>','Output source list (default: stdout)',ofn);
        n:=1000;      defparam('-n','Maximum number of sources (default: 1000)',n);
        ra:=0;        defparam('-ra','Center RA coordinate in degrees (default: 0)',ra);
        dec:=0;       defparam('-dec','Center DEC coordinate in degrees (default: 0)',dec);
        width:=1;     defparam('-x','Field width in degree (default: 1)',width);
        height:=1;    defparam('-y','Field height in degree (default: 1)',height);
        rot:=0;       defparam('-r','Clockwise rotation in degree. 0 means north is up. (Default: 1)',rot);
        scale:=0;     defparam('-s','Scale in arcsecs/pixel (Default: 1)',scale);
        proj:=-1;     defparam('-p','Projection method. -1: gnomonic (rectilinear), 0: stereographic, 1: orthographic. Default: -1.',proj);
        year:=-1;     defparam('-a','Year for proper motion calculation (default: previous year)',year);
                      defparam('-o','Output source list',ofn);    deffmt('-o','<outfile>');
                      defparam('-h','Help',helpf);
        run;
        done;
    end;
    if scale<0.1 then scale:=0.1;

    if year<0 then year:=yearof(now)-1;
    cc.debug('ppmsources: n='+int2str(n)+' out='+ofn+' dbd='+dbd+' year='+int2str(year)+' ra='+real2str(ra)+' dec='+real2str(dec)+' width='+real2str(width)+' height='+real2str(height)+' rot='+real2str(rot)+' scale='+real2str(scale)+' proj='+real2str(proj));

    { sources }
    pl.pt:=nil;
    mi:=CSImg_MetaInfo.create();
    cc.Info('ppmsources: Calculating sources');
    SImg_PPMSources(cc, dbd, year, ra,dec,width,height,rot,scale, proj, n, pl,mi);
    cc.Info('getsources: Writing source data array '+ofn);
    SImg_WFITS(cc,CSImg_FileOutput.init(ofn),pl,mi).close(cc);

    SImg_Free(cc,pl);
    mi.destroy;
    cc.destroy;
end.
