{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name test}
{@description Continuum subtraction with interpolation of bright stars.}

{$I directives.pas}
uses simage,getparam,sysutils;

procedure bgapprox_test(cc:CsImg_Config);
const nxy = 1000;
      nc  = 3;
      sig = 1 * 4;
var fi,fj,fe    : TSImg_FPImage;
    x,y,c       : longint;
    m           : array[1..nc] of longint;
    d           : ^simgfp;
begin
//    randomize;
    for c:=1 to nc do
        m[c]:=round(nxy*(1-exp(ln(0.5)*c)));
    fi.pt:=nil;  SImg_New(cc, fi, nxy,nxy,nc, false);
    d:=fi.pt;
    for y:=0 to nxy-1 do
        for x:=0 to nxy-1 do
            for c:=1 to nc do
            begin
                if (x<m[c]) and (y<m[c]) then d^:=1+(random-0.5)*sig
                else d^:=(random-0.5)*sig;
            d+=1;
            end;
    SImg_WriteImageFile(cc, 'bgApproxTest-fi.fits',fi);

    fj.pt:=nil; fe.pt:=nil;
    SImg_Duplicate(cc, fi,fj, true);
    SImg_BGApprox3(cc, fj,fe,fe,fe, 2,false);
    for x:=0 to fj.size-1 do
        fj.pt[x]:=fi.pt[x]-fj.pt[x];
    SImg_WriteImageFile(cc, 'bgApproxTest-fg.fits',fj);
    SImg_Free(cc, fj);

//    SImg_Approx(cc, fi,1,1,0, 0,25, false, 3);
//    SImg_WriteImageFile(cc, 'approxTest-fg.fits',fi);
    SImg_Approx(cc, fi,1,1,0, 0,10, true, 3);
    SImg_WriteImageFile(cc, 'approxTest-fj.fits',fi);

    SImg_Free(cc, fi);
end;

procedure colorNoise_test(cc:CsImg_Config);
const nx  = 1400;
      ny  = 400;
      nf  = 127*SQRT(0.08);
var rd          : TSImg_RDImage;
    fi          : TSImg_FPImage;
    i,j         : longint;
    r           : array[0..2] of simgfp;
begin
    randomize;
    rd.pt:=nil; fi.pt:=nil;
    SImg_Noise_Norm(cc, fi, nx,ny,3);
    SImg_filter(cc, fi,fi, 1);
    SImg_New(cc, rd, SImg_RawByte, nx,ny,3, false);
    r[0]:=nf/sqrt(0.21);
    r[1]:=nf/sqrt(0.71);
    r[2]:=nf/sqrt(0.08);
    for i:=0 to fi.size-1 do
    begin
        j:=round(127+fi.pt[i]*r[i mod 3]);
        if j<0 then j:=0;
        if j>255 then j:=255;
        rd.ptb[i]:=j;
    end;
    SImg_WriteImageFile(cc, 'colorNoiseOpt.ppm',rd);

    r[0]:=nf*sqrt(3);
    for i:=0 to fi.size-1 do
    begin
        j:=round(127+fi.pt[i]*r[0]);
        if j<0 then j:=0;
        if j>255 then j:=255;
        rd.ptb[i]:=j;
    end;
    SImg_WriteImageFile(cc, 'colorNoiseEq.ppm',rd);
    SImg_Free(cc, rd);
end;



var cc                  : CSImg_Config;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);

//    SImg_SubtractStars_PSFFitTest(cc);
//    SImg_StarSubtract_WidthFilter_tests(cc);
//    bgapprox_test(cc);
    colorNoise_test(cc);

    cc.destroy;
end.
