{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{*************************************************************************}
{******* PNM (image format) routines *************************************}
{*************************************************************************}
{*************************************************************************}

{*************************************************************************}
{******* SImg_RPNM *******************************************************}
{*************************************************************************}
//define[@readstr(][);][
    #0:='';
    if d>=fb.size then exit;
    while (d<fb.size) and (fb.pt[d]<>32) and (fb.pt[d]<>9) and (fb.pt[d]<>10) and (fb.pt[d]<>13) and (fb.pt[d]<>35) do
    begin
        #0+=fb.ptc[d];
        inc(d);
    end;
    repeat
        if (d<fb.size) and (fb.pt[d]=35) then
        begin
            e:=d;
            repeat
                inc(d);
            until (d>=fb.size) or (fb.pt[d]=10) or (fb.pt[d]=13);
            if (d>e+12) and (fb.pt[e+1]=32) and (fb.pt[e+10]=32) and (fb.pt[e+11]=61) then readkey(e+2, e+12, d-1);
        end;
        inc(d);
    until ((d>=fb.size)) or ((fb.pt[d]<>32) and (fb.pt[d]<>9) and (fb.pt[d]<>10) and (fb.pt[d]<>13) and (fb.pt[d]<>35));
]

//define[@readlaststr(][);][
    #0:='';
    while (d<fb.size) and (fb.pt[d]<>32) and (fb.pt[d]<>9) and (fb.pt[d]<>10) and (fb.pt[d]<>13) do
    begin
        #0+=fb.ptc[d];
        inc(d);
    end;
    inc(d);
{    if (d<fb.size) and (fb.pt[d]=35) then
    repeat
        inc(d);
    until (d>=fb.size) or (fb.pt[d]=10) or (fb.pt[d]=13); }
    if d>fb.size then exit;
]

//define[@val(][,#1,#2);][
    val(#0,#1,#2);
    if #2<>0 then exit;
]

function SImg_RPNM(cc:CSImg_Config; fb:TSImg_FileBuf; var ri:TSImg_RDImage; mi:CSImg_MetaInfo):boolean;
var s,t       : string31;
    i,j,d,e   : longint;
    cf,co     : double;
    fmt       : TSImg_RawFormat;
    bole,bobe : boolean;

    procedure readkey(ka,va,vz:longint);
    var k       : string15;
        v,c     : shortstring;
        i       : longint;
        r       : double;
        q       : boolean;
    begin
        if mi=nil then exit;
        // key
        move(fb.pt[ka],k[1],8);
        k[0]:=#8;
        trimstr(k);
        // comment
        i:=va;
        q:=false;
        while (i<=vz) and (q or (fb.pt[i]<>47)) do
        begin
            if fb.pt[i]=39 then q:=not q;
            inc(i);
        end;
        if vz>i+255 then vz:=i+255;
        c[0]:=#0;
        if i<vz then
        begin
            move(fb.pt[i+1],c[1],vz-i);
            c[0]:=chr(vz-i);
            trimstr(c);
        end;
        // value
        if i>va+255 then i:=va+255;
        if va<i then move(fb.pt[va],v[1],i-va);
        v[0]:=chr(i-va);
        trimstr(v);
        // processing
        mi.add(k,v,c);
        upstr(k);
        if k='BZERO' then
        begin
            val(v,r,i);
            if i=0 then co:=r;
        end;
        if k='BSCALE' then
        begin
            val(v,r,i);
            if i=0 then cf:=r;
        end;
        if k='BYTEORDR' then
        begin
            if v='''BIG_ENDIAN''' then bobe:=true;
            if v='''LITTLE_ENDIAN''' then bole:=true;
        end;
    end;

begin
    if cc=nil then cc:=SImg_Config_Default;
    cc.debug('SImg_RPNM: called with: img='+SImg_InfoStr2(ri)+'   '+int2str(fb.size)+' Bytes input buffer');
    cf:=1; co:=0;
    bole:=false; bobe:=false;

    SImg_RPNM:=false;
    d:=0;
    @readstr(s);
    t:=s;

    if (t<>'P5') and (t<>'P6') and (t<>'P8') then exit;

    @readstr(s);
    @val(s,ri.nx,i);
    @readstr(s);
    @val(s,ri.ny,i);
    if t='P6' then ri.nc:=3
    else if t='P5' then ri.nc:=1
    else begin
        @readstr(s);
        @val(s,ri.nc,i);
    end;
    @readlaststr(s);
    @val(s,j,i);
    if j>255 then fmt:=SImg_RawWord
    else fmt:=SImg_RawByte;
    SImg_New(cc, ri, fmt, ri.nx, ri.ny, ri.nc,false);
    ri.cf:=cf; ri.co:=co;
    ri.bole:=(bole or cc.pnmLE) and (not bobe);
    move(fb.pt[d], ri.pt^, ri.dsize);
    SImg_RPNM:=true;
end;

//udefine[@val(]
//udefine[@readstr(]
//udefine[@readlaststr(]

{*************************************************************************}
{******* SImg_WPNM *******************************************************}
{*************************************************************************}
//define[@append(][);][
    if length(#0)+buflen>bufsize then
    begin
        f.write(cc,@buf,buflen);
        buflen:=0;
    end;
    move(#0[1],buf[buflen],length(#0));
    buflen+=length(#0);
]

procedure SImg_WPNM(cc:CSImg_Config; f:CSImg_Output; ri:TSImg_RDImage; mi:CSImg_MetaInfo);
const bufsize = 16*1024;
var buf           : array[0..bufsize-1] of byte;
    s             : shortstring;
    rj            : TSImg_RDImage;
    bole          : boolean;
    buflen,i      : longint;
begin
    cc.debug('SImg_WPNM: called with: output='+f.info+' ri='+SImg_InfoStr2(ri));
    if not SImg_Recover(cc,ri) then cc.Error('SImg_WPNM: Invalid image');
    buflen:=0;

    if (ri.fmt=SImg_RawByte) or (ri.fmt=SImg_RawWord) then rj:=ri
    else begin
        rj.pt:=nil;
        SImg_Convert(cc,ri,rj,SImg_RawWord,nil,true,true);
    end;

    bole:=(cc.forceLE) and (not cc.forceBE);
    if rj.bole<>bole then SImg_SwapBytes(rj);

    f.open(cc);
    case ri.nc of
        1 : s:='P5'#10;
        3 : s:='P6'#10;
        else s:='P8'#10;
    end;
    s+='# BZERO    = '+real2str(rj.co,15)+#10;
    s+='# BSCALE   = '+real2str(rj.cf,15)+#10;
    s+='# BYTEORDR = '+condstring(bole,'''LITTLE_ENDIAN'''#10,'''BIG_ENDIAN'''#10);
    @append(s);
    if mi<>nil then
        for i:=0 to mi.size-1 do
            if mi.key[i]<>'' then
            begin
                fillchar(s,sizeof(s),#32);
                s[0]:=#10;
                s[1]:='#';
                s[12]:='=';
                move(mi.key[i][1],s[3],length(mi.key[i]));
                if (mi.val[i]<>'') then s+=' = ' + mi.val[i];
                if (mi.comment[i]<>'') and (length(s)<252) then
                begin
                    if length(s)<35 then s[0]:=#35;
                    s+=' / '+mi.comment[i];
                end;
                s+=#10;
                @append(s);
            end;
    s:=int2str(ri.nx)+' '+int2str(ri.ny)+#10;
    if (ri.nc<>1) and (ri.nc<>3) then s+=int2str(ri.nc)+#10;
    if rj.fmt=SImg_RawByte then s+='255'#10
    else s+='65535'#10;
    @append(s);
    f.write(cc, @buf, buflen);
    f.write(cc, rj.pt, rj.dsize);
    f.close(cc);
    if rj.pt<>ri.pt then SImg_FRee(cc,rj);
end;

//udefine[@append(]
