{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{******* SImg_Upscale ****************************************************}
{*************************************************************************}
threadprocedure SImg_Upscale_thread(fi:TSImg_FPImage; fk:TSImg_FPImage; y1:longint; y2:longint; nx:longint; ny:longint);
var rx,ry,r00,r01,r10,r11,xr,yr :  double;
    x,y,c,xi,yi,ya,yb           : longint;
    d,e                         : psimgfp;
    algo,b                      : pbyte;
begin
    rx:=(fi.nx-1.001)/(nx-1);
    ry:=(fi.ny-1.001)/(ny-1);

    ya:=trunc(y1*ry);
    yb:=trunc(y2*ry);
    getmem2(algo,(yb-ya+1)*fi.nx);

    for y:=ya to yb do
    begin
        d:=fi.pt + y*fi.sl;
        b:=algo + (y-ya)*fi.nx;
        for x:=0 to fi.nx-2 do
        begin
            r00:=0; r01:=0; r10:=0; r11:=0;
            for c:=0 to fi.nc-1 do
            begin
                r00+=d[0];
                r01+=d[fi.nc];
                r10+=d[fi.sl];
                r11+=d[fi.sl+fi.nc];
                d+=1;
            end;
            if abs(r11-r00)>abs(r01-r10)
            then b^:=1
            else b^:=0;
            b+=1;
        end;
    end;

    e:=fk.pt+y1*fk.sl;
    for y:=y1 to y2 do
    begin
        yi:=trunc(y*ry);
        yr:=y*ry-yi;
        for x:=0 to nx-1 do
        begin
            xi:=trunc(x*rx);
            xr:=x*rx-xi;
            c:=algo[(yi-ya)*fi.nx + xi];
            if c=1 then
            begin
                if xr+yr<=1 then
                begin
                    r00:=1-xr-yr;
                    r01:=xr;
                    r10:=yr;
                    r11:=0;
                end
                else begin
                    r11:=xr+yr-1;
                    r10:=1-xr;
                    r01:=1-yr;
                    r00:=0;
                end;
            end
            else begin
                if xr<=yr then
                begin
                    r10:=yr-xr;
                    r11:=xr;
                    r00:=1-yr;
                    r01:=0;
                end
                else begin
                    r01:=xr-yr;
                    r00:=1-xr;
                    r11:=yr;
                    r10:=0;
                end;
            end;
            d:=fi.pt + yi*fi.sl + xi*fi.nc;
            for c:=0 to fi.nc-1 do
            begin
                e^:=d[0]*r00 + d[fi.nc]*r01 + d[fi.sl]*r10 + d[fi.sl+fi.nc]*r11;
                e+=1;
                d+=1;
            end;
        end;
    end;

    freemem2(algo);
end;

{ nx and ny should be larger than fi.nx and fi.ny, respectively }
procedure SImg_Upscale(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; nx,ny:longint);
var fk    : TSImg_FPImage;
    i     : longint;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Upscale: fi='+SImg_InfoStr2(fi)+' fj='+SImg_InfoStr2(fj)+' ny='+int2str(nx)+' ny='+int2str(ny));
    if not SImg_Recover(cc,fi) then cc.error('Invalid image: fi');
    if not SImg_OverwriteAlert(fi,fj) then SImg_Free(cc,fj);
    fk.pt:=nil;
    SImg_New(cc,fk,nx,ny,fi.nc);

    for i:=0 to cc.threads-1 do
        dothread SImg_Upscale_thread(fi,fk, (i*ny) div cc.threads, (((i+1)*ny) div cc.threads)-1, nx,ny);
    threadwait;

    SImg_Move(cc, fk,fj);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_Resize *****************************************************}
{*************************************************************************}
{ ti,tn : thread index, number
  fi,fj : input,output image
  sx,sy : scaling factors (>1 means downscale)
  nx,ny : sub-sampling grid size, samples 1..nx-1 and 1..ny-1 where 0,nx,ny lie on the pixel boundary
}
threadprocedure SImg_Resize_mean(ti:longint; tn:longint; fi:TSImg_FPImage; fj:TSImg_FPImage; sx:double; sy:double; nx:longint; ny:longint);
var buf                 : array of double;
    d,e                 : ^simgfp;
    x,y,c,xo,yo,cnt     : longint;
    ya,yz,xj,yj,yi,xi   : longint;
    r,dx,dy             : double;
begin
    ya:=(fj.ny*ti) div tn;
    yz:=(fj.ny*(ti+1)) div tn-1;
    dx:=sx/nx;
    dy:=sy/ny;
    nx-=1;
    ny-=1;
    xo:=round((fi.nx-fj.nx*sx)*0.5); { aligns boundary of first pixel }
    yo:=round((fi.ny-fj.ny*sy)*0.5);
    setlength(buf,fj.nc);
    for c:=0 to fj.nc-1 do
        buf[c]:=0;
    e:=fj.pt + ya*fj.sl;
    r:=1/(nx*ny);
    for yj:=ya to yz do
        for xj:=0 to fj.nx-1 do
        begin
            cnt:=0;
            for y:=1 to ny do
            begin
                yi:=trunc(yj*sy + y*dy + yo);
                if (yi<0) or (yi>=fi.ny) then continue;
                for x:=1 to nx do
                begin
                    xi:=trunc(xj*sx + x*dx + xo);
                    if (xi<0) or (xi>=fi.nx) then continue;
                    cnt+=1;
                    d:=fi.pt + yi*fi.sl + xi*fi.nc;
                    for c:=0 to fj.nc-1 do
                        buf[c]+=d[c];
                end;
            end;
            if cnt>0 then r:=1/cnt
            else r:=0;  { should not happen }
            for c:=0 to fj.nc-1 do
            begin
                e[c]:=buf[c]*r;
                buf[c]:=0;
            end;
            e+=fj.nc;
        end;
    setlength(buf,0);
end;

threadprocedure SImg_Resize_rms(ti:longint; tn:longint; fi:TSImg_FPImage; fj:TSImg_FPImage; sx:double; sy:double; nx:longint; ny:longint);
var buf                 : array of double;
    d,e                 : ^simgfp;
    x,y,c,xo,yo,cnt     : longint;
    ya,yz,xj,yj,yi,xi   : longint;
    r,dx,dy             : double;
begin
    ya:=(fj.ny*ti) div tn;
    yz:=(fj.ny*(ti+1)) div tn-1;
    dx:=sx/nx;
    dy:=sy/ny;
    nx-=1;
    ny-=1;
    xo:=round((fi.nx-fj.nx*sx)*0.5); { aligns boundary of first pixel }
    yo:=round((fi.ny-fj.ny*sy)*0.5);
    setlength(buf,fj.nc);
    for c:=0 to fj.nc-1 do
        buf[c]:=0;
    e:=fj.pt + ya*fj.sl;
    r:=1/(nx*ny);
    for yj:=ya to yz do
        for xj:=0 to fj.nx-1 do
        begin
            cnt:=0;
            for y:=1 to ny do
            begin
                yi:=trunc(yj*sy + y*dy + yo);
                if (yi<0) or (yi>=fi.ny) then continue;
                for x:=1 to nx do
                begin
                    xi:=trunc(xj*sx + x*dx + xo);
                    if (xi<0) or (xi>=fi.nx) then continue;
                    cnt+=1;
                    d:=fi.pt + yi*fi.sl + xi*fi.nc;
                    for c:=0 to fj.nc-1 do
                        buf[c]+=sqr(d[c]);
                end;
            end;
            if cnt>0 then r:=1/cnt
            else r:=0;  { should not happen }
            for c:=0 to fj.nc-1 do
            begin
                e[c]:=sqrt(buf[c]*r);
                buf[c]:=0;
            end;
            e+=fj.nc;
        end;
    setlength(buf,0);
end;

threadprocedure SImg_Resize_quant(ti:longint; tn:longint; fi:TSImg_FPImage; fj:TSImg_FPImage; sx:double; sy:double; nx:longint; ny:longint; q:double);
var buf                 : array of simgfp;
    d,e                 : ^simgfp;
    x,y,c,xo,yo,cnt     : longint;
    ya,yz,xj,yj,yi,xi   : longint;
    m,j,k,k0            : longint;
    r,dx,dy,qf          : simgfp;
begin
    ya:=(fj.ny*ti) div tn;
    yz:=(fj.ny*(ti+1)) div tn-1;
    dx:=sx/nx;
    dy:=sy/ny;
    nx-=1;
    ny-=1;
    qf:=1;
    if q>0.5 then
    begin
        q:=1-q;
        qf:=-1;
    end;
    m:=trunc(nx*ny*q+0.05)+1;
    setlength(buf, m*fj.nc);
    for c:=0 to m*fj.nc-1 do
        buf[c]:=0; { must be a valid number }
    xo:=round((fi.nx-fj.nx*sx)*0.5); { aligns boundary of first pixel }
    yo:=round((fi.ny-fj.ny*sy)*0.5);
    e:=fj.pt + ya*fj.sl;
    r:=1/(nx*ny);
    for yj:=ya to yz do
        for xj:=0 to fj.nx-1 do
        begin
            cnt:=0;
            for y:=1 to ny do
            begin
                yi:=trunc(yj*sy + y*dy + yo);
                if (yi<0) or (yi>=fi.ny) then continue;
                for x:=1 to nx do
                begin
                    xi:=trunc(xj*sx + x*dx + xo);
                    if (xi<0) or (xi>=fi.nx) then continue;
                    d:=fi.pt + yi*fi.sl + xi*fi.nc;
                    if cnt<m then j:=cnt
                    else j:=m-1;
                    cnt+=1;
                    for c:=0 to fj.nc-1 do
                    begin
                        r:=d[c]*qf;
                        k0:=c*m;
                        k:=k0+j;
                        if (cnt>m) and (r>buf[k]) then continue;
                        while (k>k0) and (r<buf[k-1]) do
                        begin
                            buf[k]:=buf[k-1];
                            k-=1;
                        end;
                        buf[k]:=r;
                    end;
                end;
            end;
            j:=trunc(cnt*q-0.05);
            k:=trunc(cnt*q+0.05);
            for c:=0 to fj.nc-1 do
                e[c]:=0.5*(buf[c*m+j]+buf[c*m+k])*qf;
            e+=fj.nc;
        end;
    setlength(buf, 0);
end;

{ fi    : input
  fo    : output
  sx,sy : scaling factors, >1: down-scaling
  filt  : 0..1: quantile; <0 mean; >1 RMS
}
procedure SImg_Resize(cc:CSImg_Config; fi:TSImg_FPImage; var fo:TSImg_FPImage; sx,sy:double; filt:double);
var nx,ny,i     : longint;
    fj          : TSImg_FPImage;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Resize: fi='+SImg_InfoStr2(fi)+' fo='+SImg_InfoStr2(fo)+' sx='+real2str(sx)+' sy='+real2str(sx)+' filt='+real2str(filt));
    if not SImg_Recover(cc,fi) then cc.error('Invalid image: fi');
    if not SImg_OverwriteAlert(fi,fo) then SImg_Free(cc,fo);
    if abs(sx)<0.001 then
    begin
        if sx<0 then sx:=-0.001
        else sx:=0.001;
    end;
    if abs(sy)<0.001 then
    begin
        if sy<0 then sy:=-0.001
        else sy:=0.001;
    end;

    { sub sampling grid in fj pixels }
    if sx<1 then nx:=3
    else nx:=round(sx*2+1.4);
    if sy<1 then ny:=3
    else ny:=round(sy*2+1.4);

    fj.pt:=nil;
    SImg_New(cc, fj, round(fi.nx/abs(sx)), round(fi.ny/abs(sy)), fi.nc, false);
    for i:=0 to cc.threads-1 do
        if filt<=1e-5 then
        begin
            dothread SImg_resize_mean(i, cc.threads, fi,fj, sx,sy, nx,ny);
        end
        else if filt>=1+1e-5 then
        begin
            dothread SImg_resize_rms(i, cc.threads, fi,fj, sx,sy, nx,ny);
        end
        else begin
            dothread SImg_resize_quant(i, cc.threads, fi,fj, sx,sy, nx,ny, filt);
        end;
    threadwait;

    SImg_Move(cc, fj,fo);
    SImg_DecDebugLevel;
end;

