{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

// define[ppm_test]
//define[sources_debug]

const SImg_sources_maxChainLength  = 32;
      SImg_sources_mindist         = 20;        // minimum distance between sources
      SImg_sources_areaaDiv        = 8;         // divide images in SImg_sources_areaaDiv x SImg_sources_areaDiv tiles

{*******************************************************************************}
{********* SImg_GetSources *****************************************************}
{*******************************************************************************}
type TSImg_source_int=record
       x,y      : longint;   // position of brightest pixel (5x5 LP-filtered)
       xc,yc    : single;    // center relative to 1st estimate
       b        : double;    // value of brightest pixel  (5x5 LP-filtered)
       ba       : double;    // brightness (integrated flux, above background)
       g        : double;    // background
       w        : double;    // weight
       v        : boolean;   // valid
       c        : boolean;   // need to be re-checked for double occurrences
       end;
     PSImg_source_int = ^TSImg_source_int;

     TSImg_source_int2=record  // used for finding brighter neighbor sources
       x,y      : double;    // position of brightest pixel (5x5 LP-filtered)
       b        : double;    // value of brightest pixel  (5x5 LP-filtered)
       g        : double;    // background
       w        : double;    // weight
       xd,yd  : array[0..SImg_sources_maxChainLength-1] of double;
       idx    : array[0..SImg_sources_maxChainLength-1] of word;
       len    : longint;
       end;
     PSImg_source_int2 = ^TSImg_source_int2;

     TSImg_source_int3=record  // used for finding brighter neighbor sources
       x,y    : double;    // position of brightest pixel (5x5 LP-filtered)
       b      : double;    // value of brightest pixel  (5x5 LP-filtered)
       end;
     PSImg_source_int3 = ^TSImg_source_int3;


{********* SImg_GetSources_thread **********************************************}
{ get position and brightness of sources
    fj          : source
    fw          : pixel weight, assumed as 1/sqr(sigma)
    xr,yr       : region (0..SImg_sources_areaaDiv-1 , 0..SImg_sources_areaaDiv-1)
    sourcei     : internal source table
    rad         : distance between sources
    sat         : saturation limit
    minw        : weight threshold. If weight <0 pixel is marked as invalid. Ignored if fw.pt=nil
}
threadprocedure SImg_GetSources_thread(cc:CSImg_Config; fj:TSImg_FPImage; fw:TSImg_FPImage; xr:longint; yr:longint; sourcei:PSImg_source_int; n2:longint; rad:double; sat:double; minw:double );
var xa,ya,xz,yz   : longint;
    x,y,r2        : longint;
    i,j,k         : longint;
    a,r,s,t,u,v,w : double;
    d,e           : ^simgfp;
    valid         : boolean;

label l1,l2;

begin
    if cc=nil then cc:=SImg_Config_Default;
    if (not SImg_Recover(cc,fj)) or (fj.nc<>1) then cc.error('SImg_GetSources_int: invalid image');
    xa:=2+(((fj.nx-4)*xr) div SImg_sources_areaaDiv);
    xz:=1+(((fj.nx-4)*(xr+1)) div SImg_sources_areaaDiv);
    ya:=2+(((fj.ny-4)*yr) div SImg_sources_areaaDiv);
    yz:=1+(((fj.ny-4)*(yr+1)) div SImg_sources_areaaDiv);

    { calculate background }
    r:=0; a:=1e200;
    for y:=ya to yz do
    begin
        d:=fj.pt+y*fj.sl+xa;
        for x:=xa to xz do
        begin
            r+=d^;
            if d^<a then a:=d^;
            d+=1;
        end;
    end;
    r/=(yz-ya+1)*(xz-xa+1);
    r+=r-a;
    j:=0; a:=0;
    for y:=ya to yz do
    begin
        d:=fj.pt+y*fj.sl+xa;
        for x:=xa to xz do
        begin
            if d^<r then
            begin
                a+=d^;
                j+=1;
            end;
            d+=1;
        end;
    end;
    a:=a/j*25;  { 5x5 pixel average }

    { find sources }
    for i:=0 to n2-1 do
    begin
        sourcei[i].v:=false;
        sourcei[i].c:=false;
        sourcei[i].ba:=-1;
        sourcei[i].x:=123;
        sourcei[i].y:=456;
    end;
    r2:=trunc(sqr(rad)+0.01);
    r:=0;
    for k:=0 to 1 do    { two iterations because sources are unsorted }
    begin
        for y:=ya to yz do
        begin
            d:=fj.pt + (y-2)*fj.sl + xa-2;
            for x:=xa to xz do
            begin
                e:=d; d+=1;
                s:=-a;
                for i:=0 to 4 do
                begin
                    s+=e[0]+e[1]+e[2]+e[3]+e[4];
                    e+=fj.sl;
                end;
                if s<=r then goto l1; { fast decision }
                { slow decision }
                j:=-1;
                for i:=0 to n2-1 do
                    if (sourcei[i].v) and (sqr(sourcei[i].x-x)+sqr(sourcei[i].y-y)<=r2) then
                    begin
                        if s<=sourcei[i].b then goto l1;
                        sourcei[i].v:=false;
                        j:=i;
                    end;
                { insert new source }
                if j>=0 then i:=j
                else begin
                    i:=0; j:=0;
                    while (i<n2) and sourcei[i].v do
                    begin
                        if sourcei[i].b<sourcei[j].b then j:=i;
                        i+=1;
                    end;
                    if i>=n2 then i:=j;
                end;
                sourcei[i].x:=x;
                sourcei[i].y:=y;
                sourcei[i].b:=s;
                sourcei[i].v:=true;
                { get values for fast decision }
                r:=1e200;
                for i:=0 to n2-1 do
                begin
                    if not sourcei[i].v then
                    begin
                        r:=0;
                        continue;
                    end;
                    if sourcei[i].b<r then r:=sourcei[i].b;
                end;
l1:
            end;
        end;
    end;

    { remove sources near outer border }
    if (xa<8) or (xz>=fj.nx-8) or (ya<8) or (yz>=fj.ny-8) then
        for i:=0 to n2-1 do
            with sourcei[i] do
                if v and (x<9) or (x>=fj.nx-9) or (y<9) or (y>=fj.ny-9) then v:=false;

    { mark sources near inner border for re-check }
    i:=round(rad);
    if xr>0 then xa+=i;
    if xr<SImg_sources_areaaDiv-1 then xz-=i;
    if yr>0 then ya+=i;
    if yr<SImg_sources_areaaDiv-1 then yz-=i;
    for i:=0 to n2-1 do
        with sourcei[i] do
            if v and ((x>=xz) or (x<=xa) or (y>=yz) or (y<=ya)) then c:=true;

    { get centers and brightness }
    for i:=0 to n2-1 do
        if sourcei[i].v then
        begin
            { background + variance }
            a:=0; v:=0; k:=0;
            for y:=-9 to 9 do
                for x:=-9 to 9 do
                begin
                    j:=sqr(y)+sqr(x);
                    if (j>=50) and (j<100) then { 156 pixels }
                    begin
                        r:=fj.pt[ (sourcei[i].y+y)*fj.sl + sourcei[i].x+x];
                        a+=r;
                        v+=sqr(r);
                        k+=1;
                    end;
                end;
            a/=k; v:=v/k-sqr(a);
            { center brightness above noise }
            s:=-9*a;
            d:=fj.pt + (sourcei[i].y-1)*fj.sl + sourcei[i].x;
            for y:=-1 to 1 do
            begin
                s+=d[-1]+d[0]+d[1];
                d+=fj.sl;
            end;
            if (s<=0) or (sqr(s)<324*v) then    { s<2*9*sqrt(v) }
            begin
                 sourcei[i].v:=false;
                 goto l2;
            end;
            { brightness, center }
            r:=0; s:=0; t:=0;
            w:=0; k:=0;
            valid:=true;
            for y:=-7 to 7 do
                for x:=-7 to 7 do
                begin
                    j:=sqr(x)+sqr(y);
                    if j<50 then        { 149 pixels }
                    begin
                        u:=fj.pt[(sourcei[i].y+y)*fj.sl + sourcei[i].x+x];
                        valid:=valid and (u<sat);
                        u-=a;
                        if (j<16) and (u>0) then
                        begin
                            r+=x*u;
                            s+=y*u;
                            t+=u;
                        end;
                        w+=u;
                        k+=1;
                        if (fw.pt<>nil) and valid then
                        begin
                            d:=fw.pt + (sourcei[i].y+y)*fw.sl + (sourcei[i].x+x)*fw.nc;
                            for j:=0 to fw.nc-1 do
                                valid:=valid and (d[j]>minw);
                        end;
                    end;
                end;
            if (t>1e-100) and (abs(r)<t) and (abs(s)<t) and (w>0) and (sqr(w)>4*v*k) then  { w/sqrt(k)>2*sqrt(v) }
            begin
                t:=1/t;
                sourcei[i].xc:=r*t;
                sourcei[i].yc:=s*t;
                sourcei[i].ba:=w;
                sourcei[i].g:=a;
                if valid then sourcei[i].w:=1
                else sourcei[i].w:=0;
            end
            else sourcei[i].v:=false;
l2:
        end;
end;

{********* SImg_GetSources_neighbours ********************************************}
{ find brighter neighbors (sorted be distance and brightness)
    num         : number of sources
    sourcei     : internal source table
    t,tn        : thread, number of threads
}
threadprocedure SImg_GetSources_neighbours(num:longint; sourcei:PSImg_source_int2; t:longint; tn:longint);
var i,j,k      : longint;
    p          : PSImg_source_int2;
    xd,yd,d    : double;
begin
    for i:=0 to num-1 do
        if i mod tn = t then
        begin
            k:=0;
            p:=sourcei+i;
            for j:=i-1 downto 0 do
            begin
                xd:=sourcei[j].x-p^.x;
                yd:=sourcei[j].y-p^.y;
                d:=sqr(xd)+sqr(yd);
                while (k>0) and (d<=sqr(p^.xd[k-1])+sqr(p^.yd[k-1])) do
                    k-=1;
                if k<SImg_sources_maxChainLength then
                begin
                    p^.xd[k]:=xd;
                    p^.yd[k]:=yd;
                    p^.idx[k]:=j;
                    k+=1;
                end;
            end;
            p^.len:=k;
     end;
end;


{********* SImg_Sources_getchains *******************************************}
procedure SImg_Sources_getchains(cc:CSImg_Config; sourcei: PSImg_Source_int2; num:longint; var pl:TSImg_RDImage);
var i,j,c,tn,n2 : longint;
threadvars;
begin
    tn:=cc.threads;
    for i:=0 to num-1 do
        sourcei[i].len:=0;

    { find chain of brighter neighbors }
    for i:=0 to tn-1 do
        dothread SImg_GetSources_neighbours(num, sourcei, i,tn);
    threadwait;
    n2:=0;
    for i:=0 to num-1 do
        if sourcei[i].len>n2 then n2:=sourcei[i].len;

    SImg_New(cc, pl, SImg_RawDouble, 5+n2*3,num,1, true);
    for i:=0 to num-1 do
    begin
        j:=i*pl.sl;
        with sourcei[i] do
        begin
            pl.ptd[j+0]:=x;
            pl.ptd[j+1]:=y;
            pl.ptd[j+2]:=b;
            pl.ptd[j+3]:=g;
            pl.ptd[j+4]:=w;
            j+=5;
            for c:=0 to len-1 do
            begin
                pl.ptd[j+0]:=xd[c];
                pl.ptd[j+1]:=yd[c];
                pl.ptd[j+2]:=idx[c];
                j+=3;
            end;
            for c:=len to n2-1 do
            begin
                pl.ptd[j+0]:=0;
                pl.ptd[j+1]:=0;
                pl.ptd[j+2]:=-1;
                j+=3;
            end;
        end;
    end;
end;

{********* SImg_GetSources *****************************************************}
{ get position and brightness of sources
    fi   : input image
    fw   : weight of input image (1/sqr(noise))
    pl   : source data array
    num  : maximum number of sources
    cmp  : compress, i.e. do no save chains

  Format of source data array:
    0   : x
    1   : y
    2   : brightness (integrated flux minus background)
    3   : background (not used)
    4   : weight. Should be either 0 (is saturated) or 1
    5.. : source chain
        0 : xd
        1 : yd
        2 : index
  }
procedure SImg_GetSources(cc:CSImg_Config; fi,fw:TSImg_FPImage; var pl:TSImg_RDImage; num:longint; cmp:boolean);
var fj               : TSImg_FPImage;
    tmp              : TSImg_Source_int;
    x,y,n2,i,j,c,tn  : longint;
    sourcei          : ^TSImg_Source_int;
    sourcei2         : ^TSImg_Source_int2;
    a,rad,sat,minw   : double;
    d,e              : ^simgfp;
    threads          : PSImg_GetSources_thread;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_GetSources: fi='+SImg_InfoStr2(fi)+'  fw='+SImg_InfoStr2(fw)+'  num='+int2str(num));
    if not SImg_Recover(cc,fi) then cc.error('SImg_GetSources: Illegal input image');
    if (fi.nx<32) or (fi.ny<32) then cc.error('SImg_GetSources: Image to small');
    if num<1 then cc.Error('SImg_GetSources: num must be >0');
    if (fw.pt<>nil) and (not SImg_Recover(cc,fw)) then
    begin
        cc.warning('SImg_GetSources: illegal image: fw');
        fw.pt:=nil;
    end;
    if (fw.pt<>nil) and ((fw.nx<>fi.nx) or (fw.ny<>fi.ny)) then
    begin
        cc.warning('SImg_GetSources: invalid dimensions of image fw');
    fw.pt:=nil;
    end;

    { convert to 1 color image and get saturation limit }
    sat:=-1e200;
    if fi.nc<>1 then
    begin
        SImg_New(cc,fj,fi.nx,fi.ny,1);
        d:=fi.pt; e:=fj.pt;
        for i:=0 to fj.size-1 do
        begin
            a:=0;
            for j:=0 to fi.nc-1 do
            begin
                a+=d^; d+=1;
            end;
            e^:=a; e+=1;
            if a>sat then sat:=a;
        end;
    end
    else begin
        fj:=fi;
        for i:=0 to fj.size-1 do
        begin
            a:=fj.pt[i];
            if a>sat then sat:=a;
        end;
    end;
    sat*=0.75;

    { weight threshold }
    minw:=0;
    if fw.pt<>nil then
    begin
        for i:=0 to fw.size-1 do
            minw+=fw.pt[i];
        minw/=(fw.size*1e4);
    end;

    { get the sources }
    i:=sqr(SImg_sources_areaaDiv);
    j:=(num + num div 8) div i;
    if j<4 then j:=4;
    n2:=j*i;
    rad:=sqrt(fi.nx*fi.ny/num)*0.5;
    if rad<SImg_Sources_mindist then rad:=SImg_Sources_mindist;
    getmem2(sourcei,sizeof(TSImg_Source_int)*n2);

    tn:=cc.threads;
    getmem(threads,tn*sizeof(CSImg_GetSources_thread));
    for i:=0 to tn-1 do
        threads[i]:=CSImg_GetSources_thread.create;

    i:=0;
    for y:=0 to SImg_sources_areaaDiv-1 do
        for x:=0 to SImg_sources_areaaDiv-1 do
        begin
            threads[i mod tn].wait;
            threads[i mod tn].setwork(cc, fj,fw, x,y, sourcei+i*j,j, rad, sat,minw);
            i+=1;
        end;
    for j:=0 to tn-1 do
        threads[j].wait;
    for j:=0 to tn-1 do
        threads[j].destroy;
    freemem(threads);
    if fi.pt<>fj.pt then SImg_Free(cc,fj);

    { remove double occurrences }
    c:=trunc(sqr(rad)+0.01);
    for i:=0 to n2-1 do
        if sourcei[i].c and sourcei[i].v then
        begin
            x:=sourcei[i].x;
            y:=sourcei[i].y;
            for j:=i+1 to n2-1 do
                if sourcei[j].c and sourcei[j].v and (sqr(sourcei[j].x-x)+sqr(sourcei[j].y-y)<=c) then
                begin
                    if sourcei[j].b<sourcei[i].b then sourcei[j].v:=false
                    else sourcei[i].v:=false;
                end
        end;

    { sort by brightness }
    for i:=1 to n2-1 do
    begin
        j:=i;
        while (j>0) and sourcei[j].v and ((not sourcei[j-1].v) or (sourcei[j].ba>sourcei[j-1].ba)) do
        begin
            tmp:=sourcei[j-1];
            sourcei[j-1]:=sourcei[j];
            sourcei[j]:=tmp;
            j-=1;
        end;
    end;

    { generate output }
    while (num>0) and (not sourcei[num-1].v) do
        num-=1;
    if num<1 then cc.error('SImg_GetSources: Less than 2 sources found');

    for i:=0 to num-1 do
        with sourcei[i] do
        begin
            if not v then cc.error('getsources: Internal error 1');
            if abs(xc)>3 then cc.error('getsources: Internal error 2');
            if abs(yc)>3 then cc.error('getsources: Internal error 3');
        end;


    {for i:=num downto 0 do
        with sourcei[i] do
            writeln(i,':  ',x,'+',xc:0:2, '  ',y,'+',yc:0:2, '     ', ba:0:5,'   ',g:0:5,'   ',w);}
    if cmp then
    begin
        SImg_New(cc, pl, SImg_RawDouble, 5,num,1, true);
        for i:=0 to num-1 do
        begin
            j:=i*pl.sl;
            with sourcei[i] do
            begin
                pl.ptd[j+0]:=x+xc;
                pl.ptd[j+1]:=y+yc;
                pl.ptd[j+2]:=ba;
                pl.ptd[j+3]:=g;
                pl.ptd[j+4]:=w;
            end;
        end;
        freemem2(sourcei);
    end
    else begin
        getmem2(sourcei2,sizeof(TSImg_Source_int2)*num);
        for i:=0 to num-1 do
        begin
            sourcei2[i].x:=sourcei[i].x+sourcei[i].xc;
            sourcei2[i].y:=sourcei[i].y+sourcei[i].yc;
            sourcei2[i].b:=sourcei[i].ba;
            sourcei2[i].g:=sourcei[i].g;
            sourcei2[i].w:=sourcei[i].w;
        end;
        freemem2(sourcei);
        SImg_Sources_getchains(cc, sourcei2,num, pl);
        freemem2(sourcei2);
    end;

    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_PlotSources ****************************************************}
{*******************************************************************************}
{ plot fposition and brightness of sources
    fo  : output
    pl  : source data array
    mi  : keys image boundaries IBND* must be set
    rad : half brightness radius
    sc  : downscale by this factor
}
procedure SImg_PlotSources(cc:CSImg_Config; var fo:TSImg_FPImage; pl:TSImg_RDImage; mi:CSImg_MetaInfo; rad,sc:double);
var fj                     : TSImg_FPImage;
    x,y,i,j,r2,xa,xz,ya,yz : longint;
    ymin,ymax,xmin,xmax    : longint;
    x0,y0                  : longint;
    r,sig,xc,yc            : double;
    psf                    : ^single;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_PlotSources:  fo='+SImg_InfoStr2(fo)+' pl='+SImg_InfoStr2(pl)+' rad='+real2str(rad)+' sc='+real2str(sc));
    if not (SImg_Recover(cc,pl) and (pl.fmt=SImg_RawDouble) and (pl.nx>=5) and (pl.nc=1)) then cc.Error('SImg_PlotSources: Invalid format of source data array');
    if not mi.get('IBNDXMIN', r) then cc.error('SImg_PlotSources: Attribute not found: "IBNDXMIN"'); xmin:=round(r);
    if not mi.get('IBNDXMAX', r) then cc.error('SImg_PlotSources: Attribute not found: "IBNDXMAX"'); xmax:=round(r);
    if not mi.get('IBNDYMIN', r) then cc.error('SImg_PlotSources: Attribute not found: "IBNDYMIN"'); ymin:=round(r);
    if not mi.get('IBNDYMAX', r) then cc.error('SImg_PlotSources: Attribute not found: "IBNDYMAX"'); ymax:=round(r);
    if sc<simg_mina then sc:=1
    else sc:=1/sc;

    SImg_New(cc, fj, round((xmax-xmin+1)*sc), round((ymax-ymin+1)*sc), 1);
    r2:=round(rad*8); j:=2*r2+1;
    sig:=-ln(2)/sqr(rad);
    getmem2(psf,sizeof(single)*sqr(j));
    for y:=-r2 to r2 do
        for x:=-r2 to r2 do
            psf[(y+r2)*j+x+r2]:=exp((sqr(x)+sqr(y))*sig);

    for i:=0 to pl.ny-1 do
    begin
        xc:=pl.ptd[i*pl.sl+0];
        yc:=pl.ptd[i*pl.sl+1];
        r :=pl.ptd[i*pl.sl+2];
        x:=round((xc-xmin)*sc);
        y:=round((yc-ymin)*sc);
        xa:=x-r2;  if xa<0 then xa:=0;
        xz:=x+r2;  if xz>=fj.nx then xz:=fj.nx-1;
        ya:=y-r2;  if ya<0 then ya:=0;
        yz:=y+r2;  if yz>=fj.ny then yz:=fj.ny-1;
        x0:=x-r2;  y0:=y-r2;
        for y:=ya to yz do
            for x:=xa to xz do
                fj.pt[y*fj.sl + x] += psf[(y-y0)*j+x-x0]*r;
    end;

    freemem2(psf);

    SImg_Move(cc, fj,fo);
    SImg_DecDebugLevel;
end;

{*******************************************************************************}
{********* SImg_PPMSources *****************************************************}
{*******************************************************************************}
{ extract sources from ppm database
    dbd   : databse directory
    year  : year for proper motion calculation
    ra,dec,width,height  : extraction area, all parameters in degree
    rot   : rotation extraction area, cw, 0 mean north is up
    scale : pixel scale in arcsecs
    proj  : projection type: -1: gnomonic; 0: stereographic, 1: orthographic
    n     : maximum number of sources
    pl    : source data array
    mi    : mata data (optional)
}
procedure SImg_PPMSources(cc:CSImg_Config; const dbd:ansistring; year:longint; ra,dec,width,height,rot,scale, proj:double; n:longint; var pl:TSImg_RDImage; mi:CSImg_MetaInfo);
const deg2rad = pi/180;
      rad2deg = 180/pi;
      r2pi    = 1/(2*pi);
      mag_ref = 15;     // reference magnitude for brightness 1
var indexer             : CPPMDB_Indexer;
    tiles               : array of CPPMDB_Tile;
    tile                : CPPMDB_Tile;
    fn                  : ansistring;
    sourcei3            : ^TSImg_Source_int3;
    sourcei2            : ^TSImg_Source_int2;
    xi,yi,nx,ny,tn      : longint;
    ra_i,dec_i,i,j      : longint;
    mag,mag_n,sn,sn2    : longint;
    x,y,z,r,sdec,cdec   : double;
    srot,crot,maxra     : double;
    p1,p2,a,b,srad      : double;
//ifdef[ppm_test]
    wcs                 : TSImg_WCS;
    tx,ty,tra,tdec      : double;
    mag0,mag1,mag2      : longint;
//endif

begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_PPMSources: dbd='+dbd+' year='+int2str(year)+' n='+int2str(n)+' ra='+real2str(ra)+' dec='+real2str(dec)+' width='+real2str(width)+' height='+real2str(height)+' rot='+real2str(rot)+' scale='+real2str(scale)+' proj='+real2str(proj));
    indexer:=CPPMDB_Indexer.create(dbd);
    if n<10 then n:=10;
    rot*=deg2rad;
    ra*=deg2rad;
    dec*=deg2rad;
    srot:=sin(rot);
    crot:=cos(rot);
    sdec:=sin(dec);
    cdec:=cos(dec);
    { projection }
    p1:=1; p2:=1;
    if proj<-1 then p1:=0
    else if proj<0 then p1:=proj+1
    else if proj<1 then p2:=1-proj
    else p2:=0;
    { number of test grid points, see below }
    nx:=round(2.5*width/180*indexer.dec_n+0.5001);
    ny:=round(2.5*height/180*indexer.dec_n+0.5001);
    { scale: factor for sphere coordinates to pixel }
    if scale<0.01 then scale:=0.01;
    scale:=1/(pi*scale/(180*3600));
    { width/height in pixel }
    if width>180 then width:=180;
    if height>180 then height:=180;
    width*=0.5*deg2rad;
    height*=0.5*deg2rad;
    x:=p1+p2*cos(width);
    if x<0.25 then x:=0.25;
    y:=p1+p2*cos(height);
    if y<0.25 then y:=0.25;
    width :=sin(width )*(p1+p2)/x*scale;
    height:=sin(height)*(p1+p2)/y*scale;
    { Determine the tiles need. This is achieved by projecting points of a grid into an equatorial coordinate system.
      The test grid size is 40% of the tile size.  }
    setlength(tiles,16);
    tn:=0;
    maxra:=0;
    for yi:=0 to ny do
        for xi:=0 to nx do
        begin
            x:=(2*xi/nx-1)*width;
            y:=(2*yi/ny-1)*height;
//ifdef[ppm_test]
            cc.info('xi='+int2str(xi)+' yi='+int2str(yi)+'  -->  xpix='+real2str(x,6)+' ypix='+real2str(y,6));
//endif
            { rotation }
            r:=x;
            x:=crot*x + srot*y;
            y:=crot*y - srot*r;
            { scaling + sphere coordinates }
            x/=scale; y/=scale;
            a:=(sqr(x)+sqr(y))*sqr(p2) + sqr(p1+p2);
            b:=p1*(p1+p2);
            r:=sqr(b)+(sqr(p2)-sqr(p1))*a;
            if r<=-1e6 then continue;   // test point is not a projection
            if r<=0 then r:=0           // may happen due to rounding errors
            else r:=sqrt(r);
            r:=(b+r)/a;
            x*=r; y*=r;
            z:=1-sqr(x)-sqr(y);
            if z<0 then continue;       // test point is not a projection
            z:=sqrt(z);
//ifdef[ppm_test]
            cc.info('xi='+int2str(xi)+' yi='+int2str(yi)+'  -->  x='+real2str(x,6)+' y='+real2str(y,6)+' z='+real2str(z,6));
//endif
            { rotate by DEC towards north pole }
            r:=z;
            z:=cdec*z + sdec*y;
            y:=cdec*y - sdec*r;
            { calculate RA,DEC }
            y:=-arcsin(y);
            x:=-arctan2(x,z);  { RA, DEC }
            if abs(x)>maxra then maxra:=abs(x);
            x+=ra;
            indexer.radec_i(x*rad2deg,y*rad2deg, ra_i,dec_i);
            i:=0;
            while (i<tn) and ((tiles[i].ra_i<>ra_i) or (tiles[i].dec_i<>dec_i)) do
                i+=1;
            if i>=tn then
            begin
                if i>high(tiles) then setlength(tiles,i+16);
                fn:=dbd+pathDelim+ppmdb_fileBase(ra_i,dec_i)+'.pdb';
                cc.info('ppmsources: reading ppm database tile tile '+fn);
                tiles[i]:=CPPMDB_Tile.create(fn);
                tn+=1;
            end;

//ifdef[ppm_test]
            cc.info('xi='+int2str(xi)+' yi='+int2str(yi)+'  -->  ra='+real2str(x*rad2deg,-8)+' dec='+real2str(y*rad2deg,-8));
            // reverse transform
            x-=ra;
            { transform coordinates }
            { sphere coordinate }
            r:=  cos(y);  {z}
            y:= -sin(y);
            z:= r*cos(x);
            x:=-r*sin(x);
            { rotate by DEC towards field center }
            r:=z;
            z:=cdec*z - sdec*y;
            y:=cdec*y + sdec*r;
            { projection + scaling }
            //if z<1e-3 then continue;
            r:=scale*(p1+p2)/(p1+p2*z);
            x*=r; y*=r;
            { rotation }
            r:=x;
            x:=crot*x - srot*y;
            y:=crot*y + srot*r;
            //if abs(x)>width then continue;
            //if abs(y)>height then continue;
            cc.info('xi='+int2str(xi)+' yi='+int2str(yi)+'  -->  xpix='+real2str(x,6)+' ypix='+real2str(y,6));
            cc.info('');
//endif
        end;

    mag_n:=tiles[0].mag_n;
    { get a valid tiles[].dbuf  }
    for i:=0 to tn-1 do
    begin
        tiles[i].decode(year);
        if tiles[i].mag_n<>mag_n then cc.error('SImg_PPMSources: PPM database tiles have different magnitude settings');
    end;
    a:=-ln(2.511886)/mag_n;    { get sources  }
    getmem2(sourcei3,sizeof(TSImg_Source_int)*n);
    for i:=0 to n-1 do
    begin
        sourcei3[i].x:=0;
        sourcei3[i].y:=0;
        sourcei3[i].b:=0;
    end;

    srad:=trunc(width*height/n+0.01);
    if srad<5 then srad:=5;
    mag:=0; sn:=0;
    while (sn<n) and (mag<32767) do
    begin
        sn2:=sn;
        for i:=0 to tn-1 do
        begin
            tile:=tiles[i];
            while (tile.dbuf.m[0]<=mag) and (tile.dbuf.m[0]>=0) and (sn<n) do
            begin
//ifdef[ppm_test]
                if sn=0 then
                begin
                    tra:=tile.dbuf.ra;
                    tdec:=tile.dbuf.dec;
                    mag0:=tile.dbuf.m[0];
                    mag1:=tile.dbuf.m[1];
                    mag2:=tile.dbuf.m[2];
                end;
//endif
                x:=tile.dbuf.ra*deg2rad-ra;
                x:=x-round(x*r2pi)*2*pi;
                y:=tile.dbuf.dec*deg2rad;
                b:=tile.dbuf.m[0]-(2+mag_ref)*mag_n;
                tile.decode(year);
                if abs(x)>maxra then continue;
                { transform coordinates }
                { sphere coordinate }
                r:=  cos(y);  { =z }
                y:= -sin(y);
                z:= r*cos(x);
                x:=-r*sin(x);
                { rotate by DEC towards field center }
                r:=z;
                z:=cdec*z - sdec*y;
                y:=cdec*y + sdec*r;
                { projection + scaling }
                if z<1e-30 then continue;
                r:=scale*(p1+p2)/(p1+p2*z);
                x*=r; y*=r;
                { field rotation }
                r:=x;
                x:=crot*x - srot*y;
                y:=crot*y + srot*r;
                if abs(x)>width then continue;
                if abs(y)>height then continue;
                { search for close neighbors }
                j:=0;
                repeat
                    r:=sqr(sourcei3[j].x-x)+sqr(sourcei3[j].y-y);
                    j+=1;
                until (r<srad) or (j>=sn);
                if (r>1) and (r<srad) then continue;
//ifdef[ppm_test]
                if sn=0 then
                begin
                    tx:=x;
                    ty:=y;
                    cc.info('brightest source: mag='+real2str(b/mag_n+mag_ref)+' x='+real2str(tx,-1)+' y='+real2str(ty,-1)+'  mag0='+real2str(mag0/mag_n-2,-1)+'  mag1='+real2str(mag1/mag_n-2,-1)+'  mag2='+real2str(mag2/mag_n-2,-1));
                end;
//endif
                { magnitude }
                b:=exp(a*b);
                { save source }
                if r<srad then
                begin { multiple stars }
                    j-=1;
                    z:=sourcei3[j].b;
                    sourcei3[j].x:=(sourcei3[j].x*z + x*b)/(z+b);
                    sourcei3[j].y:=(sourcei3[j].y*z + y*b)/(z+b);
                    sourcei3[j].b+=b;
                end else
                begin
                    sourcei3[sn].x:=x;
                    sourcei3[sn].y:=y;
                    sourcei3[sn].b:=b;
                    sn+=1;
                end;
            end;
        end;
        mag+=1;
    end;
    if (sn>=n) and (sn2>10) then
    begin
        sn:=sn2;  // equal density preferred
        mag-=1;
    end;
    for i:=0 to tn-1 do
        tiles[i].destroy;
    setlength(tiles,0);
    indexer.destroy;
    cc.info('ppmsources: faintest source: mag='+real2str((mag-1)/mag_n-1,-1));

    { generate output }
    if sn<1 then cc.error('SImg_PPMSources: Less than 2 sources found');
    getmem2(sourcei2,sizeof(TSImg_Source_int2)*sn);
    for i:=0 to sn-1 do
    begin
        sourcei2[i].x:=sourcei3[i].x;
        sourcei2[i].y:=sourcei3[i].y;
        sourcei2[i].b:=sourcei3[i].b;
        sourcei2[i].g:=0;
        sourcei2[i].w:=1;
    end;
    freemem2(sourcei3);
    SImg_Sources_getchains(cc, sourcei2,sn, pl);
    freemem2(sourcei2);

    { metadata }
    if mi<>nil then
    begin
        xi:=round(width+0.4);
        yi:=round(height+0.4);
        i:=round(proj);
        a:=180/(pi*scale);
        if i<0 then fn:='TAN'
        else if i>0 then fn:='SIN'
        else fn:='STG';
        mi.add('IBNDXMIN',int2str(-xi),'Lower image boundary in X direction');
        mi.add('IBNDXMAX',int2str( xi),'Upper image boundary in X direction');
        mi.add('IBNDYMIN',int2str(-yi),'Upper image boundary in Y direction');
        mi.add('IBNDYMAX',int2str( yi),'Lower image boundary in Y direction');
        { wcs data, cordnate -width, height becomes (1,1)  }
        mi.add('CRVAL1', real2str(ra*rad2deg,15),'RA coordinate of reference pixel');
        mi.add('CUNIT1', 'deg', 'RA unit');
        mi.add('CTYPE1', 'RA---'+fn, 'Axis and projection');
        mi.add('CRVAL2', real2str(dec*rad2deg,15),'DEC coordinate of reference pixel');
        mi.add('CUNIT2', 'deg', 'DEC unit');
        mi.add('CTYPE2', 'DEC--'+fn, 'Axis and projection');
        mi.add('CRPIX1', int2str(xi+1),'X coordinate of reference pixel');
        mi.add('CRPIX2', int2str(yi+1),'Y coordinate of reference pixel');
        mi.add('CD1_1', real2str(-crot*a,15),'Rotation and scaling matrix');
        mi.add('CD1_2', real2str( srot*a,15),'Rotation and scaling matrix');
        mi.add('CD2_1', real2str( srot*a,15),'Rotation and scaling matrix');
        mi.add('CD2_2', real2str( crot*a,15),'Rotation and scaling matrix');
//ifdef[ppm_test]
        SImg_GetWCSData(cc, mi, wcs);
        for j:=0 to 1 do
            for i:=0 to 1 do
            begin
                x:=width*(2*i-1);
                y:=height*(2*j-1);
                SImg_GetRaDec(cc, wcs, x+xi, y+yi, 2*yi+1, x,y);
                cc.info(int2str(i)+','+int2str(j)+': ra='+real2str(x,-8)+' dec='+real2str(y,-8));
            end;
        tx+=xi;  ty+=yi;
        SImg_GetRaDec(cc, wcs, tx,ty,2*yi+1, x,y);
        cc.info('ra='+real2str(tra,12)+' dec='+real2str(tdec,12)+'  -->  x='+real2str(tx,6)+' y='+real2str(ty,6)+'  -->');
        cc.info('ra='+real2str(x,12)+' dec='+real2str(y,12));
//endif
    end;
    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_GetSourceChains ************************************************}
{*******************************************************************************}
{ find chains of brighter neigbours
    pli   : input source data array
    plo   : output source data array
    mi    : mata data (optional)
}
procedure SImg_GetSourceChains(cc:CSImg_Config; pli:TSImg_RDImage; var plo:TSImg_RDImage);
var sourcei     : ^TSImg_Source_int2;
    i,j         : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_GetSourceChain: pli='+SImg_InfoStr2(pli)+' plo='+SImg_InfoStr2(plo));
    getmem2(sourcei,sizeof(TSImg_Source_int2)*pli.ny);
    for i:=0 to pli.ny-1 do
    begin
        j:=i*pli.nx;
        with sourcei[i] do
        begin
            x:=pli.ptd[j+0];
            y:=pli.ptd[j+1];
            b:=pli.ptd[j+2];
            g:=pli.ptd[j+3];
            w:=pli.ptd[j+4];
        end;
    end;
    SImg_Sources_getchains(cc, sourcei,pli.ny, plo);
    freemem2(sourcei,sizeof(TSImg_Source_int2)*pli.ny);
    SImg_DecDebugLevel;
end;


