{  simg -- visualization library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{ ************************************************************************* }
{ ******** TSivDraw ******************************************************* }
{ ************************************************************************* }
procedure siv_draw_size_request(widget:PGtkWidget; requisition: PGtkRequisition); cdecl; forward;
procedure siv_draw_size_allocate(widget:PGtkWidget; allocation:PGdkRectangle); cdecl; forward;
function siv_draw_expose_event(widget:PGtkWidget; event:PGdkEventExpose):boolean32; cdecl; forward;
procedure siv_draw_unrealize(widget:PGtkWidget); cdecl; forward;
function siv_draw_marker_idle(draw:PSIVDraw):longint; cdecl; forward;

{ ******** siv_draw_class_init ******************************************** }
procedure siv_draw_class_init(klass: PSivDrawClass); cdecl;
var widget_class  : PGtkWidgetClass absolute klass;
begin
    klass^.old_unrealize:=widget_class^.unrealize;
    klass^.old_size_allocate:=widget_class^.size_allocate;
    widget_class^.size_request:=@siv_draw_size_request;
    widget_class^.size_allocate:=@siv_draw_size_allocate;
    widget_class^.expose_event:=@siv_draw_expose_event;
    widget_class^.unrealize:=@siv_draw_unrealize;
end;

{ ******** siv_draw_init ************************************************** }
procedure siv_draw_init(draw: PSivDraw); cdecl;
begin
    with draw^ do
    begin
        rgb:=nil;
        rgb_free:=nil;
        image_width:=0;
        image_height:=0;
        hadj:=nil;
        vadj:=nil;
        rgb_width:=0;
        rgb_height:=0;
        rgb_x0:=0;
        rgb_y0:=0;
        image:=nil;
        marker_x:=-1;
        marker_y:=-1;
        marker_time:=stimer-siv_marker_update_time;
        background_gc:=nil;
        shadow_type:=GTK_SHADOW_IN;
        extra_width:=0;
        extra_height:=0;
        buf:=nil;
        back_buf:=nil;
        last_width:=-1;
        last_height:=-1;
        last_x:=0;
        last_y:=0;
        left:=0; right:=0;
        top:=0; bottom:=0;
    end;
end;

{ ******** siv_draw_get_type ********************************************** }
const siv_draw_info : TGtkTypeInfo =
         ( Type_Name : 'SivDraw';
            object_size : sizeof(TSivDraw);
            class_size : sizeof(TSivDrawClass);
            class_init_func : TGtkClassInitFunc(@siv_draw_class_init);
            object_init_func : TGtkObjectInitFunc(@siv_draw_init);
            reserved_1 : nil;
            reserved_2 : nil;
            base_class_init_func : nil;
         );

const siv_type_draw : TGtkType = 0;

function siv_draw_get_type:dword;  cdecl;
begin
    if siv_type_draw=0 then siv_type_draw:=gtk_type_unique(gtk_drawing_area_get_type, @siv_draw_info);
    result:=siv_type_draw;
end;

{ ******** siv_is_draw_class ********************************************** }
function  siv_is_draw_class(klass:pointer):boolean;
begin
    siv_is_draw_class:=(klass<>nil) and (PGtkTypeClass(klass)^.g_type=siv_draw_get_type);
end;

{ ******** siv_si_draw **************************************************** }
function  siv_is_draw(obj:pointer):boolean;
begin
    siv_is_draw:=(obj<>nil) and siv_is_draw_class(PGtkTypeObject(obj)^.g_class);
end;

{ ******** siv_draw_size_request ****************************************** }
procedure siv_draw_size_request(widget:PGtkWidget; requisition: PGtkRequisition); cdecl;
var draw  : PSIVDraw absolute widget;
begin
    if widget=nil then
    begin
        siv_warning('siv_draw_size_request: widget=nil');
        exit;
    end;
    if not siv_is_draw(draw) then
    begin
        siv_warning('siv_draw_size_request: PSivDraw expected');
        exit;
    end;
    if requisition=nil then
    begin
        siv_warning('siv_draw_size_request: requisition=nil');
        exit;
    end;
    with draw^ do
    begin
        if hadj=nil then requisition^.width:=image_width+extra_width*2
        else if requisition^.width<extra_width*2+10 then requisition^.width:=extra_width*2+20;
        if vadj=nil then requisition^.height:=image_height+extra_height*2
        else if requisition^.height<extra_height*2+10 then requisition^.height:=extra_height*2+20;
    end;
    {writeln('draw.size_request: ',ew,'  ',requisition^.width,'  ',requisition^.height);}
end;

{ ******** siv_draw_configure ********************************************** }
procedure siv_draw_configure(widget:PGtkWidget; allocation: PGdkRectangle);
const extra_render = 70;
var draw         : PSivDraw absolute widget;
    dw,dh,dx,dy  : longint;
    xa,ya,xz,yz  : longint;
begin
    if widget=nil then
    begin
        siv_warning('siv_draw_configure: widget=nil');
        exit;
    end;
    if not siv_is_draw(draw) then
    begin
        siv_warning('siv_draw_configure: PSivDraw expected');
        exit;
    end;
    if allocation=nil then
    begin
        siv_warning('siv_draw_configure: allocation=nil');
        exit;
    end;

    with draw^ do
    begin
        if (allocation^.width<1) or (allocation^.height<1) then exit;

        dw:=allocation^.width-2*extra_width;
        dx:=round((image_width-dw)*siv_adjustment_get_pos(hadj));
        dh:=allocation^.height-2*extra_width;
        dy:=round((image_height-dh)*siv_adjustment_get_pos(vadj));

//       writeln('draw.size_allocate.1: ',dx,'  ',dy,'  ',dw,'  ',dh);
        if (dw=last_width) and (dh=last_height) and (dx=last_x) and (dy=last_y) then exit;

        if (dw<>last_width) or (dh<>last_height) then
        begin
            if buf<>nil then gdk_pixmap_unref(buf);
            buf:=gdk_pixmap_new(widget^.window, allocation^.width, allocation^.height, -1);
            gdk_draw_rectangle(buf, PGtkStyle(widget^.style)^.bg_gc[GTK_STATE_NORMAL], 1, 0, 0, allocation^.width, allocation^.height);
            if background_gc<>nil then gdk_draw_rectangle(buf, background_gc, 1, 0, 0, allocation^.width, allocation^.height);
            if shadow_type<>GTK_SHADOW_NONE then gtk_draw_shadow(PGtkStyle(widget^.style), buf, GTK_STATE_NORMAL, shadow_type, 0,0, allocation^.width, allocation^.height);
        end;

        { (check) update }
        xa:=dx;    if xa<=0 then xa:=0;
        xz:=dx+dw; if xz>image_width then xz:=image_width;
        ya:=dy;    if ya<=0 then ya:=0;
        yz:=dy+dh; if yz>image_height then yz:=image_height;
        if (image<>nil) and ((rgb=nil) or (xa<rgb_x0) or (ya<rgb_y0) or (xz>rgb_x0+rgb_width) or (yz>rgb_y0+rgb_height)) then
        begin
            rgb_x0:=xa-extra_render;
            rgb_y0:=ya-extra_render;
            rgb_width:=xz-xa+2*extra_render;
            rgb_height:=yz-ya+2*extra_render;

            if (rgb<>nil) and (rgb_free=rgb) then freemem2(rgb);
            image.getRGB(rgb, rgb_x0,rgb_y0, rgb_width, rgb_height);
            rgb_free:=rgb;
//            writeln('draw.size_allocate.2: ',rgb_x0,'  ',rgb_y0,'  ',rgb_width,'  ',rgb_height);
        end;

        { draw data }
        if rgb<>nil then
        begin
            xa:=dx;    if xa<rgb_x0 then xa:=rgb_x0;
            xz:=dx+dw; if xz>rgb_x0+rgb_width then xz:=rgb_x0+rgb_width;
            ya:=dy;    if ya<rgb_y0 then ya:=rgb_y0;
            yz:=dy+dh; if yz>rgb_y0+rgb_height then yz:=rgb_y0+rgb_height;

            if dx<xa then left:=xa-dx+extra_width
            else left:=extra_width;
            if dy<ya then top:=ya-dy+extra_height
            else top:=extra_height;
            right:=left+xz-xa;
            bottom:=top+yz-ya;
//              writeln('draw.size_allocate.3: ',xa,'  ',ya,'  ',xz,'  ',yz);
            if (xz>xa) and (yz>ya) then gdk_draw_rgb_image(buf, PGtkStyle(widget^.style)^.fg_gc[gtk_widget_state(widget)], left, top ,xz-xa, yz-ya, GDK_RGB_DITHER_NONE, @rgb[(ya-rgb_y0)*rgb_width*3+(xa-rgb_x0)*3], rgb_width*3);
{             siv_draw_marker_move(draw,marker_x,marker_y,false); }
        end
        else begin
            left:=(allocation^.width shr 1) + extra_width;
            top:=(allocation^.height shr 1) + extra_height;
            right:=left;
            bottom:=top;
        end;

        if back_buf<>nil then gdk_pixmap_unref(back_buf);
        gdk_pixmap_ref(buf);
        back_buf:=buf;
        last_width:=dw;
        last_height:=dh;
        last_x:=dx;
        last_y:=dy;
    end;
end;

{ ******** siv_draw_size_allocate  ***************************************** }
procedure siv_draw_size_allocate(widget:PGtkWidget; allocation: PGdkRectangle); cdecl;
begin
    if widget=nil then
    begin
        siv_warning('siv_draw_size_allocate: widget=nil');
        exit;
    end;
    PSivDrawClass(GTK_OBJECT_GET_CLASS(widget))^.old_size_allocate(widget,allocation);
    siv_draw_configure(widget,allocation);
end;

{ ******** siv_draw_adjustment_changed *********************************** }
function siv_draw_adjustment_changed(widget,draw:PGtkWidget):boolean32; cdecl;
begin
    result:=false;
    if draw=nil then
    begin
        siv_warning('siv_draw_configure: draw=nil');
        exit;
    end;
    if not siv_is_draw(draw) then
    begin
        siv_warning('siv_draw_configure: PSivDraw expected');
        exit;
    end;

    siv_draw_configure(draw,@(draw^.allocation));
    gtk_widget_queue_draw(draw);
end;

{ ******** siv_draw_expose_event ****************************************** }
function siv_draw_expose_event(widget:PGtkWidget; event:PGdkEventExpose):boolean32; cdecl;
var draw   : PSivDraw absolute widget;
begin
    result:=false;
    if widget=nil then
    begin
        siv_warning('siv_draw_expose_event: widget=nil');
        exit;
    end;
    if not siv_is_draw(draw) then
    begin
        siv_warning('siv_draw_expose_event: PSivDraw expected');
        exit;
    end;
    if event=nil then
    begin
        siv_warning('siv_draw_expose_event: event=nil');
        exit;
    end;
//    writeln('draw.expose: ',event^.area.x,'  ', event^.area.y,'  ', event^.area.width,'  ', event^.area.height);
    if draw^.buf<>nil then gdk_draw_pixmap(widget^.window, PGtkStyle(widget^.style)^.fg_gc[gtk_widget_state(widget)], draw^.buf,
            event^.area.x, event^.area.y,
            event^.area.x, event^.area.y, event^.area.width, event^.area.height);
    if draw^.back_buf<>nil then gdk_window_set_back_pixmap(widget^.window,draw^.back_buf,false);
    result:=false;
end;

{ ******** siv_draw_unrealize ********************************************* }
procedure siv_draw_unrealize(widget:PGtkWidget); cdecl;
var draw    : PSivDraw absolute widget;
begin
    siv_debug_begin('siv_draw_unrealize');
    if widget=nil then
    begin
        siv_warning('siv_draw_unrealize: widget=nil');
        exit;
    end;
    if not siv_is_draw(draw) then
    begin
        siv_warning('siv_draw_unrealize: PSivDraw expected');
        exit;
    end;
    g_source_remove(draw^.marker_id);
    PSivDrawClass(GTK_OBJECT_GET_CLASS(widget))^.old_unrealize(widget);
    with draw^ do
    begin
        if background_gc<>nil then gdk_gc_unref(background_gc);
        if buf<>nil then gdk_pixmap_unref(buf);
        if back_buf<>nil then gdk_pixmap_unref(back_buf);
        if (rgb<>nil) and (rgb_free=rgb) then freemem2(rgb);
    end;
    siv_debug_end;
end;

{ ******** siv_draw_new *************************************************** }
function siv_draw_new(data:pbyte; width,height:longint; bg:PGdkGC; shadow:TGtkShadowType; extra_space:longint):PGtkWidget;
var draw   : PSivDraw;
    widget : PGtkWidget absolute draw;
begin
    siv_debug_begin('siv_draw_new: data='+ptr2str(data)+' width='+int2str(width)+' height='+int2str(height)+' shadow='+int2str(shadow)+' space='+int2str(extra_space));
    draw:=PSivDraw(gtk_type_new(siv_draw_get_type));
    with draw^ do
    begin
        rgb:=data;
        rgb_width:=width;
        rgb_height:=height;
        image_width:=width;
        image_height:=height;
        rgb_x0:=0;
        rgb_y0:=0;
        background_gc:=bg;
        if background_gc<>nil then gdk_gc_ref(background_gc);
        shadow_type:=shadow;
        extra_width:=extra_space;
        extra_height:=extra_space;
        if shadow_type<>GTK_SHADOW_NONE then
        begin
            extra_width+=PGtkStyle(widget^.style)^.xthickness;
            extra_height+=PGtkStyle(widget^.style)^.ythickness;
        end;
        marker_id:=g_timeout_add(siv_marker_update_time,TGtkFunction(@siv_draw_marker_idle),widget);
        gtk_widget_set_events(widget,GDK_ENTER_NOTIFY_MASK or GDK_LEAVE_NOTIFY_MASK);
    end;
    result:=widget;
    siv_debug_end;
end;

function siv_draw_new(img:CSIV_Image; bg:PGdkGC; shadow:TGtkShadowType; extra_space:longint):PGtkWidget;  // initialize with dynamic data
var draw   : PSivDraw;
    widget : PGtkWidget absolute draw;
begin
    widget:=siv_draw_new(nil,0,0, bg,shadow, extra_space);
    draw^.image:=img;
    draw^.image_width:=draw^.image.scaledWidth;
    draw^.image_height:=draw^.image.scaledHeight;
    result:=widget;
end;

{ ******** siv_draw_set_hadj ********************************************** }
procedure siv_draw_set_hadj(draw:PSivDraw; hadj:PGtkAdjustment);
begin
    draw^.hadj:=hadj;
    g_signal_connect(GTK_OBJECT(hadj),'changed',TGtkSignalFunc(@siv_draw_adjustment_changed),draw);
    g_signal_connect(GTK_OBJECT(hadj),'value_changed',TGtkSignalFunc(@siv_draw_adjustment_changed),draw);
end;

{ ******** siv_draw_set_vadj ********************************************** }
procedure siv_draw_set_vadj(draw:PSivDraw; vadj:PGtkAdjustment);
begin
    draw^.vadj:=vadj;
    g_signal_connect(GTK_OBJECT(vadj),'changed',TGtkSignalFunc(@siv_draw_adjustment_changed),draw);
    g_signal_connect(GTK_OBJECT(vadj),'value_changed',TGtkSignalFunc(@siv_draw_adjustment_changed),draw);
end;

{ ******** siv_draw_reset ************************************************* }
procedure siv_draw_reset(draw:PSivDraw; data:pbyte; width,height:longint);
begin
    if draw=nil then
    begin
        siv_warning('siv_draw_reset: draw=nil');
        exit;
    end;
    if not siv_is_draw(draw) then
    begin
        siv_warning('siv_draw_reset: PSivDraw expected');
        exit;
    end;
    draw^.last_width:=-1;
    draw^.last_height:=-1;
    draw^.rgb:=data;
    draw^.rgb_width:=width;
    draw^.rgb_height:=height;
    draw^.image_width:=width;
    draw^.image_height:=height;
    draw^.rgb_x0:=0;
    draw^.rgb_y0:=0;
    siv_draw_configure(PGtkWidget(draw),@(PGtkWidget(draw)^.allocation));
    gtk_widget_queue_draw(PGtkWidget(draw));
end;

procedure siv_draw_reset(draw:PSivDraw);
begin
    if draw=nil then
    begin
        siv_warning('siv_draw_reset: widget=nil');
        exit;
    end;
    if not siv_is_draw(draw) then
    begin
        siv_warning('siv_draw_reset: PSivDraw expected');
        exit;
    end;
    if draw^.image=nil then
    begin
        siv_warning('siv_draw_reset: no dynamic data avalable');
        exit;
    end;
    if (draw^.rgb<>nil) and (draw^.rgb=draw^.rgb_free) then freemem2(draw^.rgb);
    draw^.last_width:=-1;
    draw^.last_height:=-1;
    draw^.rgb:=nil;
    draw^.rgb_width:=0;
    draw^.rgb_height:=0;
    draw^.image_width:=draw^.image.scaledWidth;
    draw^.image_height:=draw^.image.scaledHeight;
    draw^.rgb_x0:=0;
    draw^.rgb_y0:=0;
//    siv_draw_configure(PGtkWidget(draw),@(PGtkWidget(draw)^.allocation));
//    gtk_widget_queue_draw(PGtkWidget(draw));
    gtk_widget_queue_resize(PGtkWidget(draw));
end;

{ ******** siv_draw_redraw ************************************************ }
procedure siv_draw_redraw(draw:PSivDraw);
begin
    if draw=nil then
    begin
        siv_warning('siv_draw_redraw: widget=nil');
        exit;
    end;
    if not siv_is_draw(draw) then
    begin
        siv_warning('siv_draw_redraw: PSivDraw expected');
        exit;
    end;
    draw^.last_width:=-1;
    draw^.last_height:=-1;
//    siv_draw_configure(PGtkWidget(draw),@(PGtkWidget(draw)^.allocation));
//   gtk_widget_queue_draw(PGtkWidget(draw));
    gtk_widget_queue_resize(PGtkWidget(draw));
end;

{ ******** marker routines ************************************************ }
{ ******** siv_draw_crop_to_screen **************************************** }
procedure siv_draw_crop_to_screen(draw:PSIVDraw; var x1,y1,x2,y2:longint);
begin
    with draw^ do
    begin
        x1-=last_x; x2-=last_x;
        y1-=last_y; y2-=last_y;
        if x1<0 then x1:=0;
        if y1<0 then y1:=0;
        if x2>last_width then x2:=last_width;
        if y2>last_height then y2:=last_height;
    end;
end;

procedure siv_draw_crop_to_screen(draw:PSIVDraw; var x1,y1,x2,y2, rgb_x,rgb_y:longint);
var x,y : longint;
begin
    siv_draw_crop_to_screen(draw, x1,y1,x2,y2);
    with draw^ do
    begin
        x:=x1+last_x-rgb_x0;
        y:=y1+last_y-rgb_y0;
        if x<0 then
        begin
            x1-=x;
            x:=0;
        end;
        if y<0 then
        begin
            y1-=y;
            y:=0;
        end;
        rgb_x:=x;
        rgb_y:=y;

        x:=x2+last_x-rgb_x0-rgb_width;
        y:=y2+last_y-rgb_y0-rgb_height;
        if x>0 then x2-=x;
        if y>0 then y2-=y;
    end;
end;

{ ******** siv_draw_expose_rgb ******************************************** }
procedure siv_draw_expose_rgb(draw:PSIVDraw; x1,y1,x2,y2:longint);
var widget : PGtkWidget absolute draw;
begin
    with draw^ do
    begin
        if buf=nil then exit;
        siv_draw_crop_to_screen(draw,x1,y1,x2,y2);
        if (x2>x1) and (y2>y1) then gdk_draw_pixmap(widget^.window, PGtkStyle(widget^.style)^.fg_gc[gtk_widget_state(widget)], buf, x1+extra_width, y1+extra_height, x1+extra_width, y1+extra_height, x2-x1, y2-y1);
    end;
end;


{ ******** siv_draw_marker_expose ***************************************** }
procedure siv_draw_marker_expose(draw:PSIVDraw;x,y:longint);
var p1v,p2v  : boolean;
    x1,y1    : longint;
begin
    if draw=nil then
    begin
        siv_warning('siv_draw_marker_expose: widget=nil');
        exit;
    end;
    if not siv_is_draw(draw) then
    begin
        siv_warning('siv_draw_marker_expose: PSivDraw expected');
        exit;
    end;
    with draw^ do
    begin
        if buf=nil then exit;
        p1v:=(marker_x>=0) and (marker_y>=0) and (marker_x<image_width) and (marker_y<image_height);
        p2v:=(x>=0) and (y>=0) and (x<image_width) and (y<image_height);
        if p1v and p2v and ((abs(x-marker_x)+siv_marker1_xsize)*(abs(y-marker_y)+siv_marker1_ysize)<3*siv_marker1_xsize*siv_marker1_ysize) then
        begin
            if x<marker_x then x1:=x-(siv_marker1_xsize shr 1)
            else x1:=marker_x-(siv_marker1_xsize shr 1);
            if y<marker_y then y1:=y-(siv_marker1_ysize shr 1)
            else y1:=marker_y-(siv_marker1_ysize shr 1);
            siv_draw_expose_rgb(draw,  x1, y1, x1+abs(x-marker_x)+siv_marker1_xsize, y1+abs(y-marker_y)+siv_marker1_ysize);
        end
        else begin
            if p1v then
            begin
                x1:=marker_x-(siv_marker1_xsize shr 1);
                y1:=marker_y-(siv_marker1_ysize shr 1);
                siv_draw_expose_rgb(draw, x1, y1, x1+siv_marker1_xsize, y1+siv_marker1_ysize);
            end;
            if p2v then
            begin
                x1:=x-(siv_marker1_xsize shr 1);
                y1:=y-(siv_marker1_ysize shr 1);
                siv_draw_expose_rgb(draw, x1, y1, x1+siv_marker1_xsize, y1+siv_marker1_ysize);
            end;
        end;
    end;
end;

{ ******** siv_draw_marker_move ******************************************* }
procedure siv_draw_marker_move(draw:PSIVDraw; x,y:longint; expose:boolean);
var widget            : PGtkWidget absolute draw;
    x1,y1,x2,y2,xo,yo : longint;
    rgb_x,rgb_y       : longint;
    i,j,l             : longint;
    p1v,p2v           : boolean;
    mb                : ^byte;
    d,e,f             : ^byte;
    dd                : ^dword absolute d;
    ed                : ^dword absolute e;

begin
    if draw=nil then
    begin
        siv_warning('siv_draw_marker_move_timed: widget=nil');
        exit;
    end;
    if not siv_is_draw(draw) then
    begin
        siv_warning('siv_draw_marker_move_timed: PSivDraw expected');
        exit;
    end;
    with draw^ do
    begin
        if buf=nil then
        begin
            marker_x:=x;
            marker_y:=y;
            exit;
        end;
        p1v:=(marker_x>=0) and (marker_y>=0) and (marker_x<image_width) and (marker_y<image_height);
        p2v:=(x>=0) and (y>=0) and (x<image_width) and (y<image_height);
        if ((marker_x<>x) or (marker_y<>y)) and p1v then
        begin
            x1:=marker_x-siv_marker1_xsize shr 1;
            y1:=marker_y-siv_marker1_ysize shr 1;
            x2:=x1+siv_marker1_xsize;
            y2:=y1+siv_marker1_ysize;
            siv_draw_crop_to_screen(draw, x1,y1,x2,y2, rgb_x,rgb_y);
            if (x2>x1) and (y2>y1) then gdk_draw_rgb_image(buf, PGtkStyle(widget^.style)^.fg_gc[gtk_widget_state(widget)], x1+extra_width, y1+extra_height, x2-x1, y2-y1, GDK_RGB_DITHER_NONE, @rgb[rgb_y*rgb_width*3 + rgb_x*3], rgb_width*3);
        end;
        if p2v then
        begin
            x1:=x-siv_marker1_xsize shr 1; xo:=x1;
            y1:=y-siv_marker1_ysize shr 1; yo:=y1;
            x2:=x1+siv_marker1_xsize;
            y2:=y1+siv_marker1_ysize;
            siv_draw_crop_to_screen(draw,x1,y1,x2,y2, rgb_x,rgb_y);
            xo:=x1-xo+last_x; yo:=y1-yo+last_y;
            x2-=x1; y2-=y1;
            if (x2>0) and (y2>0) then
            begin
                getmem2(mb,x2*y2*3+4);
                marker_time:=stimer;
                l:=round(127.5+127.5*sin(marker_time*siv_marker_frequency))*$010101;
                d:=mb;
                for i:=0 to y2-1 do
                begin
                    e:=@rgb[(rgb_y+i)*rgb_width*3 + rgb_x*3];
                    f:=@siv_marker1[i+yo,xo];
                    for j:=0 to x2-1 do
                    begin
                        if f^>32 then dd^:=l
                        else dd^:=ed^;
                        d+=3; e+=3; f+=1;
                    end;
                end;
                gdk_draw_rgb_image(buf, PGtkStyle(widget^.style)^.fg_gc[gtk_widget_state(widget)], x1+extra_width, y1+extra_height, x2, y2, GDK_RGB_DITHER_NONE, @mb[0], x2*3 );
//                write(1);
                freemem2(mb);
            end;
        end;
        if expose then siv_draw_marker_expose(draw,x,y);
        marker_x:=x;
        marker_y:=y;
    end;
end;

function siv_draw_marker_idle(draw:PSIVDraw):longint; cdecl;
begin
    result:=0;
//    gdk_threads_enter;
    if draw=nil then
    begin
        siv_warning('siv_draw_marker_idle: widget=nil');
        exit;
    end;
    if not siv_is_draw(draw) then
    begin
        siv_warning('siv_draw_marker_idle: PSivDraw expected');
        exit;
    end;
    siv_draw_marker_move(draw, draw^.marker_x, draw^.marker_y, true);
    result:=1;
//    gdk_threads_leave;
end;
